module Cli exposing (..)
{-    ( Command
    , CommandErr
    , FilledCommand
    , FilledFlag
    , FilledFlagOption
    , FilledOption
    , Flag
    , FlagOption
    , Option
    , getFlag
    , getFlagOption
    , getOption
    , mkCommand
    , parseInput
    )
-}



type alias Option =
    { name : String
    }


type alias FilledOption =
    { name : String
    , value : Maybe String
    }


type alias Flag =
    { name : String
    }


type alias FilledFlag =
    { name : String
    , value : Bool
    }


type alias FlagOption =
    { name : String
    }


type alias FilledFlagOption =
    { name : String
    , value : Maybe String
    }


type alias DirtyCommand =
    Command


type alias Command =
    { name : String
    , opts : List Option
    , flags : List Flag
    , flagOpts : List FlagOption
    }


type alias FilledCommand =
    { name : String
    , opts : List FilledOption
    , flags : List FilledFlag
    , flagOpts : List FilledFlagOption
    }


type alias CommandErr =
    { command : Maybe Command
    , got : String
    , argumentErr : Maybe ArgumentErr
    }


type ArgumentErr
    = MissingArgs
    | TooManyArgs
    | UnrecognisedFlag


parseInput : List Command -> String -> Result CommandErr FilledCommand
parseInput commands input =
    let
        tokens =
            String.words input

        tail =
            case List.tail tokens of
                Just t ->
                    t

                Nothing ->
                    []
    in
    case List.head tokens of
        Just maybeCommand ->
            case List.head (List.filter (isCommand maybeCommand) commands) of
                Just command ->
                    case applyArguments command tail of
                        Ok filledCommand ->
                            Ok filledCommand

                        Err err ->
                            Err
                                { command = Just command
                                , got = input
                                , argumentErr = Just err
                                }

                Nothing ->
                    Err
                        { command = Nothing
                        , got = input
                        , argumentErr = Nothing
                        }

        Nothing ->
            Err
                { command = Nothing
                , got = input
                , argumentErr = Nothing
                }


mkCommand : Command -> Command
mkCommand command =
    { command | flags = List.append command.flags [ { name = "--help" } ] }


flagsFromTokens : List String -> List String
flagsFromTokens tokens =
    tokens
        |> List.filter isFlag


flagOptionsFromTokens : List String -> List String
flagOptionsFromTokens flags =
    flags
        |> List.filter isFlagOption


nonFlagsFromTokens : List String -> List String
nonFlagsFromTokens tokens =
    tokens
        |> List.filter (not << isFlag)
        |> List.filter (not << isFlagOption)


isFlag : String -> Bool
isFlag str =
    String.startsWith "--" str && not (String.contains "=" str)


isFlagOption : String -> Bool
isFlagOption str =
    String.startsWith "--" str && isOnlyOnce "=" str


isOnlyOnce : String -> String -> Bool
isOnlyOnce a b =
    1 == List.length (String.indices a b)


applyArguments : Command -> List String -> Result ArgumentErr FilledCommand
applyArguments command tokens =
    let
        opts =
            command.opts

        flags =
            command.flags

        flagOpts =
            command.flagOpts

        flagIns =
            flagsFromTokens tokens

        flagOptIns =
            flagOptionsFromTokens tokens

        nonFlagIns =
            nonFlagsFromTokens tokens
    in
    if List.length opts < List.length nonFlagIns then
        Err TooManyArgs

    else if List.any (\m -> not (List.member { name = m } flags)) flagIns then
        Err UnrecognisedFlag

    else if
        List.any
            (\m ->
                not
                    (List.member
                        { name =
                            case List.head (String.split "=" m) of
                                Just head ->
                                    head

                                Nothing ->
                                    ""
                        }
                        flagOpts
                    )
            )
            flagOptIns
    then
        Err UnrecognisedFlag

    else
        let
            optionInputs =
                expandList (List.map Just nonFlagIns) (List.length opts)
        in
        Ok
            { name = command.name
            , opts = List.map2 applyOption opts optionInputs
            , flags = List.map (fillFlag flagIns) flags
            , flagOpts = List.map (fillFlagOption flagOptIns) flagOpts
            }


applyOption : Option -> Maybe String -> FilledOption
applyOption option input =
    { name = option.name
    , value = input
    }


fillFlag : List String -> Flag -> FilledFlag
fillFlag flagTokens flag =
    if List.member flag.name flagTokens then
        { name = flag.name
        , value = True
        }

    else
        { name = flag.name
        , value = False
        }


fillFlagOption : List String -> FlagOption -> FilledFlagOption
fillFlagOption flagOptIns flagOption =
    let
        flag =
               case flagOptIns
                    |> List.filter (\m->String.contains flagOption.name m)
                    |> List.head
                of
                    Just head ->
                        Just <| String.dropLeft (1 + String.length flagOption.name) head
                    Nothing ->
                        Nothing
                    
    in
    { name = flagOption.name
    , value = flag
    }


expandList : List (Maybe String) -> Int -> List (Maybe String)
expandList input len =
    if List.length input >= len then
        List.take len input

    else
        expandList (List.append input [ Nothing ]) len


isCommand : String -> Command -> Bool
isCommand name command =
    if name == command.name then
        True

    else
        False


getOption : FilledCommand -> String -> FilledOption
getOption command name =
    case List.head <| List.filter (\m->m.name == name) command.opts of
        Just head ->
            head
        Nothing ->
            { name = name
            , value = Nothing
            }


getFlag : FilledCommand -> String -> FilledFlag
getFlag command name =
    case List.head <| List.filter (\m->m.name == name) command.flags of
        Just head ->
            head
        Nothing ->
            { name = name
            , value = False
            }


getFlagOption : FilledCommand -> String -> FilledFlagOption
getFlagOption command name =
    case List.head <| List.filter (\m->m.name == name) command.flagOpts of
        Just head ->
            head
        Nothing ->
            { name = name
            , value = Nothing
            }
